""" helper utilities """


from typing import Callable

from inspect import Signature, Parameter


def first_param(func: Callable):
    """Name of function's first parameter if any"""
    params = Signature.from_callable(func).parameters
    return next(params.__iter__(), None)


def sample_param(param: Parameter):
    """Sample value for parameters without ac default value"""

    if param.name == "expr":
        return "close"
    if param.name == "period":
        return 20
    raise NameError(param.name)


def sample_params(func: Callable) -> dict:
    """Dictionary of sample parameter values"""

    kwds = dict()

    signature = Signature.from_callable(func)
    parameters = list(signature.parameters.values())

    for param in parameters:
        if param.default != param.empty:
            continue
        if param.name in ("series", "prices"):
            continue
        value = sample_param(param)
        kwds[param.name] = value

    return kwds




"""
Mintalib Indicators

Indicators names are upper case.

Indicators offer a composable interface where a function is bound with its calculation parameters.
When instantiated with parameters an indicator yields a callable that can be applied to prices or series data.
Indicators support the `@` operator as syntactic sugar to apply the indicator to data.
So for example `SMA(50) @ prices` can be used to compute the 50 period simple moving average on `prices`,
insted of `SMA(50)(prices)`.

"""

# Do not edit! This file was generated by make-indicators.ipynb

from . import core
from .core import wrap_indicator
from .model import FuncIndicator

nan = float('nan')


__all__ = [
    'AVGPRICE', 'TYPPRICE', 'WCLPRICE', 'MIDPRICE', 'PRICE', 'CROSSOVER',
    'CROSSUNDER', 'FLAG', 'UPDOWN', 'SIGN', 'LOG', 'EXP', 'DIFF', 'LAG',
    'MIN', 'MAX', 'SUM', 'ROC', 'MAD', 'STDEV', 'SMA', 'EMA', 'RMA', 'WMA',
    'HMA', 'DEMA', 'TEMA', 'MA', 'RSI', 'DMI', 'ADX', 'PDI', 'MDI',
    'TRANGE', 'ATR', 'NATR', 'SAR', 'CCI', 'CMF', 'MFI', 'BOP', 'BBANDS',
    'KELTNER', 'KER', 'KAMA', 'MACD', 'PPO', 'SLOPE', 'RVALUE', 'FORECAST',
    'CURVE', 'STOCH', 'STREAK', 'EVAL'
]


@wrap_indicator(core.AVGPRICE)
def AVGPRICE():
    return FuncIndicator(core.AVGPRICE, params=dict(locals()))


@wrap_indicator(core.TYPPRICE)
def TYPPRICE():
    return FuncIndicator(core.TYPPRICE, params=dict(locals()))


@wrap_indicator(core.WCLPRICE)
def WCLPRICE():
    return FuncIndicator(core.WCLPRICE, params=dict(locals()))


@wrap_indicator(core.MIDPRICE)
def MIDPRICE():
    return FuncIndicator(core.MIDPRICE, params=dict(locals()))


@wrap_indicator(core.PRICE)
def PRICE(item: str = None):
    return FuncIndicator(core.PRICE, params=dict(locals()))


@wrap_indicator(core.CROSSOVER)
def CROSSOVER(level: float = 0.0, *, item: str = None):
    return FuncIndicator(core.CROSSOVER, params=dict(locals()))


@wrap_indicator(core.CROSSUNDER)
def CROSSUNDER(level: float = 0.0, *, item: str = None):
    return FuncIndicator(core.CROSSUNDER, params=dict(locals()))


@wrap_indicator(core.FLAG)
def FLAG(*, item: str = None):
    return FuncIndicator(core.FLAG, params=dict(locals()))


@wrap_indicator(core.UPDOWN)
def UPDOWN(up_level: float = 0.0, down_level: float = 0.0, *, item: str = None):
    return FuncIndicator(core.UPDOWN, params=dict(locals()))


@wrap_indicator(core.SIGN)
def SIGN(item: str = None):
    return FuncIndicator(core.SIGN, params=dict(locals()))


@wrap_indicator(core.LOG)
def LOG(*, item: str = None):
    return FuncIndicator(core.LOG, params=dict(locals()))


@wrap_indicator(core.EXP)
def EXP(*, item: str = None):
    return FuncIndicator(core.EXP, params=dict(locals()))


@wrap_indicator(core.DIFF)
def DIFF(period: int = 1, *, item: str = None):
    return FuncIndicator(core.DIFF, params=dict(locals()))


@wrap_indicator(core.LAG)
def LAG(period: int, *, item: str = None):
    return FuncIndicator(core.LAG, params=dict(locals()))


@wrap_indicator(core.MIN)
def MIN(period: int, *, item: str = None):
    return FuncIndicator(core.MIN, params=dict(locals()))


@wrap_indicator(core.MAX)
def MAX(period: int, *, item: str = None):
    return FuncIndicator(core.MAX, params=dict(locals()))


@wrap_indicator(core.SUM)
def SUM(period: int, *, item: str = None):
    return FuncIndicator(core.SUM, params=dict(locals()))


@wrap_indicator(core.ROC)
def ROC(period: int = 1, *, item: str = None):
    return FuncIndicator(core.ROC, params=dict(locals()))


@wrap_indicator(core.MAD)
def MAD(period: int = 20, *, item: str = None):
    return FuncIndicator(core.MAD, params=dict(locals()))


@wrap_indicator(core.STDEV)
def STDEV(period: int = 20, *, item: str = None):
    return FuncIndicator(core.STDEV, params=dict(locals()))


@wrap_indicator(core.SMA)
def SMA(period: int, *, item: str = None):
    return FuncIndicator(core.SMA, params=dict(locals()))


@wrap_indicator(core.EMA)
def EMA(period: int, *, adjust: bool = False, item: str = None):
    return FuncIndicator(core.EMA, params=dict(locals()))


@wrap_indicator(core.RMA)
def RMA(period: int, *, item: str = None):
    return FuncIndicator(core.RMA, params=dict(locals()))


@wrap_indicator(core.WMA)
def WMA(period: int, *, item: str = None):
    return FuncIndicator(core.WMA, params=dict(locals()))


@wrap_indicator(core.HMA)
def HMA(period: int, *, item: str = None):
    return FuncIndicator(core.HMA, params=dict(locals()))


@wrap_indicator(core.DEMA)
def DEMA(period: int, *, item: str = None):
    return FuncIndicator(core.DEMA, params=dict(locals()))


@wrap_indicator(core.TEMA)
def TEMA(period: int = 20, *, item: str = None):
    return FuncIndicator(core.TEMA, params=dict(locals()))


@wrap_indicator(core.MA)
def MA(period: int = 20, *, ma_type: str = None, item: str = None):
    return FuncIndicator(core.MA, params=dict(locals()))


@wrap_indicator(core.RSI)
def RSI(period: int = 14, *, item: str = None):
    return FuncIndicator(core.RSI, params=dict(locals()))


@wrap_indicator(core.DMI)
def DMI(period: int = 14):
    return FuncIndicator(core.DMI, params=dict(locals()))


@wrap_indicator(core.ADX)
def ADX(period: int = 14):
    return FuncIndicator(core.ADX, params=dict(locals()))


@wrap_indicator(core.PDI)
def PDI(period: int = 14):
    return FuncIndicator(core.PDI, params=dict(locals()))


@wrap_indicator(core.MDI)
def MDI(period: int = 14):
    return FuncIndicator(core.MDI, params=dict(locals()))


@wrap_indicator(core.TRANGE)
def TRANGE(*, log_prices: bool = False, percent: bool = False):
    return FuncIndicator(core.TRANGE, params=dict(locals()))


@wrap_indicator(core.ATR)
def ATR(period: int = 14):
    return FuncIndicator(core.ATR, params=dict(locals()))


@wrap_indicator(core.NATR)
def NATR(period: int = 14):
    return FuncIndicator(core.NATR, params=dict(locals()))


@wrap_indicator(core.SAR)
def SAR(afs: float = 0.02, maxaf: float = 0.2):
    return FuncIndicator(core.SAR, params=dict(locals()))


@wrap_indicator(core.CCI)
def CCI(period: int = 20):
    return FuncIndicator(core.CCI, params=dict(locals()))


@wrap_indicator(core.CMF)
def CMF(period: int = 20):
    return FuncIndicator(core.CMF, params=dict(locals()))


@wrap_indicator(core.MFI)
def MFI(period: int = 14):
    return FuncIndicator(core.MFI, params=dict(locals()))


@wrap_indicator(core.BOP)
def BOP(period: int = 20):
    return FuncIndicator(core.BOP, params=dict(locals()))


@wrap_indicator(core.BBANDS)
def BBANDS(period: int = 20, nbdev: float = 2.0):
    return FuncIndicator(core.BBANDS, params=dict(locals()))


@wrap_indicator(core.KELTNER)
def KELTNER(period: int = 20, nbatr: float = 2.0):
    return FuncIndicator(core.KELTNER, params=dict(locals()))


@wrap_indicator(core.KER)
def KER(period: int = 10, *, item: str = None):
    return FuncIndicator(core.KER, params=dict(locals()))


@wrap_indicator(core.KAMA)
def KAMA(period: int = 10, fastn: int = 2, slown: int = 30, *, item: str = None):
    return FuncIndicator(core.KAMA, params=dict(locals()))


@wrap_indicator(core.MACD)
def MACD(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    return FuncIndicator(core.MACD, params=dict(locals()))


@wrap_indicator(core.PPO)
def PPO(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    return FuncIndicator(core.PPO, params=dict(locals()))


@wrap_indicator(core.SLOPE)
def SLOPE(period: int = 20, *, item: str = None):
    return FuncIndicator(core.SLOPE, params=dict(locals()))


@wrap_indicator(core.RVALUE)
def RVALUE(period: int = 20, *, item: str = None):
    return FuncIndicator(core.RVALUE, params=dict(locals()))


@wrap_indicator(core.FORECAST)
def FORECAST(period: int = 20, offset: int = 0, *, item: str = None):
    return FuncIndicator(core.FORECAST, params=dict(locals()))


@wrap_indicator(core.CURVE)
def CURVE(period: int = 20, *, item: str = None):
    return FuncIndicator(core.CURVE, params=dict(locals()))


@wrap_indicator(core.STOCH)
def STOCH(period: int = 14, fastn: int = 3, slown: int = 3):
    return FuncIndicator(core.STOCH, params=dict(locals()))


@wrap_indicator(core.STREAK)
def STREAK(*, item: str = None):
    return FuncIndicator(core.STREAK, params=dict(locals()))


@wrap_indicator(core.EVAL)
def EVAL(expr: str):
    return FuncIndicator(core.EVAL, params=dict(locals()))





"""
    Alternate implementation of indicators for reference/testing purposes
    Calculations are mostly pandas based
"""

import numpy as np
import pandas as pd


def calc_roc(series, period: int = 1):
    """Rate of Change"""
    roc = series.pct_change(period, fill_method=None)
    roc = roc.replace([np.inf, -np.inf], np.nan)
    return roc


def calc_ema(series, period: int, adjust: bool = False):
    """Exponential Moving Average"""
    kwds = dict(span=period, adjust=adjust, ignore_na=True, min_periods=period)
    return series.ewm(**kwds).mean()


def calc_sma(series, period: int):
    """Simple Moving Average"""
    return series.rolling(window=period).mean()


def calc_sum(series, period: int):
    """Rolling Sum"""
    return series.rolling(window=period).sum()


def calc_min(series, period: int):
    """Rolling Minimum"""
    return series.rolling(window=period).min()


def calc_max(series, period: int):
    """Rolling Maximum"""
    return series.rolling(window=period).max()


def calc_stdev(series, period: int):
    """Standard Deviation"""
    return series.rolling(window=period).std(ddof=0)


def calc_mad(series, period: int):
    """Mean Aasolute deviation"""

    def _mad(s):
        return np.mean(np.fabs(s - np.mean(s)))

    return series.rolling(window=period).apply(_mad, raw=True)


def calc_wma(series, period: int = 10):
    """Weighted Moving Average"""
    weights = np.arange(1, period + 1, dtype=float)
    weights /= np.sum(weights)

    def average(data):
        return np.sum(data.values * weights)

    return series.rolling(period).apply(average)


def calc_rsi(series, period: int = 14):
    """Relative Strength Index"""
    kw = dict(alpha=1.0 / period, min_periods=period, adjust=True, ignore_na=True)

    diff = series.diff()
    ups = diff.clip(lower=0).ewm(**kw).mean()
    downs = diff.clip(upper=0).abs().ewm(**kw).mean()
    result = 100.0 - (100.0 / (1.0 + ups / downs))

    return result


def calc_macd(series, n1: int = 12, n2: int = 26, n3: int = 9):
    """Moving Averrage Convergence Divergence"""

    ema1 = calc_ema(series, n1)
    ema2 = calc_ema(series, n2)

    macd = ema1 - ema2
    signal = calc_ema(macd, n3)
    hist = macd - signal

    result = dict(macd=macd, macdsignal=signal, macdhist=hist)

    return pd.DataFrame(result)


def calc_ppo(series, n1: int = 12, n2: int = 26, n3: int = 9):
    """Price percentage oscillator"""

    ema1 = calc_ema(series, n1)
    ema2 = calc_ema(series, n2)

    ppo = (ema1 / ema2 - 1.0) * 100
    signal = calc_ema(ppo, n3)
    hist = ppo - signal

    result = dict(ppo=ppo, pposignal=signal, ppohist=hist)

    return pd.DataFrame(result)


def calc_slope(series, period: int = 20):
    """Slope (time linear regression)"""

    xx = np.arange(period) - (period - 1) / 2.0

    def func(xs):
        if np.any(np.isnan(xs)):
            return np.nan

        return np.polyfit(xx, xs, 1)[0]

    return series.rolling(period).apply(func, raw=True)




"""
Mintalib Functions

Function names are upper case.

Functions that accept a prices dataframe input have a first paramater called `prices`.
Functions that accept a series input have a fist parameter called `series`,
and an optional parameter `item` to specify which column to use on dataframe inputs.

All functions wrap their output to match the type of their input.
In particular the result of a function applied to a pandas series or dataframes
will have the same index as the input. 
"""

# Do not edit! This file was generated by make-functions.ipynb

from .core import (
    AVGPRICE, TYPPRICE, WCLPRICE, MIDPRICE, PRICE, CROSSOVER, CROSSUNDER,
    FLAG, UPDOWN, SIGN, LOG, EXP, DIFF, LAG, MIN, MAX, SUM, ROC, MAD,
    STDEV, SMA, EMA, RMA, WMA, HMA, DEMA, TEMA, MA, RSI, DMI, ADX, PDI,
    MDI, TRANGE, ATR, NATR, SAR, CCI, CMF, MFI, BOP, BBANDS, KELTNER, KER,
    KAMA, MACD, PPO, SLOPE, RVALUE, FORECAST, CURVE, STOCH, STREAK, EVAL
)

__all__ = [
    'AVGPRICE', 'TYPPRICE', 'WCLPRICE', 'MIDPRICE', 'PRICE', 'CROSSOVER',
    'CROSSUNDER', 'FLAG', 'UPDOWN', 'SIGN', 'LOG', 'EXP', 'DIFF', 'LAG',
    'MIN', 'MAX', 'SUM', 'ROC', 'MAD', 'STDEV', 'SMA', 'EMA', 'RMA', 'WMA',
    'HMA', 'DEMA', 'TEMA', 'MA', 'RSI', 'DMI', 'ADX', 'PDI', 'MDI',
    'TRANGE', 'ATR', 'NATR', 'SAR', 'CCI', 'CMF', 'MFI', 'BOP', 'BBANDS',
    'KELTNER', 'KER', 'KAMA', 'MACD', 'PPO', 'SLOPE', 'RVALUE', 'FORECAST',
    'CURVE', 'STOCH', 'STREAK', 'EVAL'
]



"""Model classes"""

import inspect

from abc import ABCMeta, abstractmethod
from types import MappingProxyType
from functools import cached_property

from .utils import format_partial, lazy_repr


class Indicator(metaclass=ABCMeta):
    """Abstact Base class for Indicators"""

    same_scale: bool = False

    __repr__ = lazy_repr

    @abstractmethod
    def __call__(self, data): ...

    @property
    @abstractmethod
    def input_type(self):
        """input type: wether "series" or "prices"""
        ...

    def __matmul__(self, other):
        if hasattr(other, 'map_batches'):
            if self.input_type == "series":
                return other.map_batches(self)
            raise NotImplementedError

        if callable(other):
            return CallableChain(self, other)
        
        return self(other)
        


class FuncIndicator(Indicator):
    """Function based Indicator"""


    def __init__(self, func, params: dict):
        self.func = func
        self.params = MappingProxyType(params)

    @cached_property
    def input_type(self):
        signature = inspect.signature(self.func)
        return next(iter(signature.parameters), None) 

    def __getattr__(self, name):
        return getattr(self.func, name)

    def __repr__(self):
        return format_partial(self.func, self.params)

    def __call__(self, prices):
        return self.func(prices, **self.params)
    

class CallableChain(Indicator):
    """Chain of Callables"""

    def __init__(self, *chain):
        items = []
        for item in chain:
            if isinstance(item, CallableChain):
                items.extend(item.chain)
            else:
                items.append(item)
        self.chain = tuple(items)

    @cached_property
    def input_type(self):
        if self.chain:
            return self.chain[-1].input_type

    def __repr__(self):
        return " @ ".join(repr(fn) for fn in self.chain)

    def __call__(self, data):
        for fn in reversed(self.chain):
            data = fn(data)
        return data




import pytest

from mintalib import core, functions, indicators
from mintalib.samples import sample_prices
from mintalib.testing import first_param, sample_params


def list_calcs():
    return [
        k
        for k, v in vars(core).items()
        if k.startswith(("calc_", "flag_"))
        and callable(v)
        and first_param(v) in ("prices", "series")
    ]


def list_functions():
    return [k for k, v in vars(functions).items() if k.isupper() and callable(v)]


def list_indicators():
    return [k for k, v in vars(indicators).items() if k.isupper() and callable(v)]


def test_samples():
    prices = sample_prices()
    assert prices is not None


@pytest.mark.parametrize("name", list_calcs())
def test_calc(name):
    print("test_calc", name)

    func = getattr(core, name)

    ftype = first_param(func)
    assert ftype in ("series", "prices")

    kwds = sample_params(func)

    item = "close" if ftype == "series" else None
    data = sample_prices(item=item)

    result = func(data, **kwds)
    assert result is not None


@pytest.mark.parametrize("name", list_functions())
def test_function(name):
    print("test_function", name)

    func = getattr(functions, name)

    ftype = first_param(func)
    assert ftype in ("series", "prices")

    kwds = sample_params(func)

    item = "close" if ftype == "series" else None
    data = sample_prices(item=item)

    result = func(data, **kwds)
    assert result is not None


@pytest.mark.parametrize("name", list_indicators())
def test_indicator(name):
    print("test_indicator", name)

    klass = getattr(indicators, name)
    kwds = sample_params(klass)
    func = klass(**kwds)

    assert callable(func)

    data = sample_prices()

    result = func(data)
    assert result is not None



"""Mintalib Utils"""

from inspect import Signature, Parameter


def format_partial(func, params, *, name: str = None):
    if name is None:
        name = func.__name__

    signature = Signature.from_callable(func)
    positional = True
    arguments = []

    for k, v in params.items():
        p = signature.parameters.get(k)

        if not p or p.kind not in (
            Parameter.POSITIONAL_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
        ):
            positional = False

        if positional:
            arguments.append(f"{v!r}")
        elif not p or v != p.default:
            arguments.append(f"{k!s}={v!r}")

    arguments = ", ".join(arguments)
    return "%s(%s)" % (name, arguments)


def lazy_repr(obj):
    data = obj.__dict__
    cname = obj.__class__.__qualname__

    signature = Signature.from_callable(obj.__init__)
    parameters = signature.parameters.values()
    positional = (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)

    arguments = []

    for p in parameters:
        v = data.get(p.name, p.default)
        if p.kind in positional:
            arguments.append(f"{v!r}")
        elif v != p.default:
            arguments.append(f"{p.name}={v!r}")

    arguments = ", ".join(arguments)

    return "%s(%s)" % (cname, arguments)



""" Sample prices data """

import pandas as pd

from importlib import resources

TIMEZONE = "America/New_York"
FREQUENCIES = "daily", "hourly", "minute"


def sample_prices(freq: str = "daily", *, max_bars: int = 0, item: str = None):
    """Sample prices"""

    if freq not in FREQUENCIES:
        raise ValueError(f"Invalid freq {freq!r}")

    fname = f"{freq}-prices.csv"
    path = resources.files(__name__).joinpath(fname)
    # Note that path here is a traversable not a Path object

    with path.open("r") as file:
        prices = pd.read_csv(file, index_col=0, parse_dates=True)

    if freq != "daily":
        prices.index = pd.to_datetime(prices.index, utc=True).tz_convert(TIMEZONE)

    if max_bars > 0:
        prices = prices.tail(max_bars)

    if item:
        return prices[item]

    return prices
